package org.example.world;

/*
 * This is free and unencumbered software released into the public domain.
 * Anyone is free to copy, modify, publish, use, compile, sell, or distribute this software, 
 * either in source code form or as a compiled binary, for any purpose, commercial or 
 * non-commercial, and by any means.
 * 
 * In jurisdictions that recognize copyright laws, the author or authors of this 
 * software dedicate any and all copyright interest in the software to the public domain. 
 * We make this dedication for the benefit of the public at large and to the detriment of 
 * our heirs and successors. We intend this dedication to be an overt act of relinquishment in 
 * perpetuity of all present and future rights to this software under copyright law.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * For more information, please refer to: https://unlicense.org/
*/

import java.io.Serializable;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.Transient;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.example.websecurity.XssSanitizerImpl;

/**
 * The City Entity for the World application.
 * <p>
 * This class represents the following DB Table:
 * <pre>
 * CREATE TABLE CITY (
 * 	ID INTEGER NOT NULL GENERATED ALWAYS AS IDENTITY (START WITH 1, INCREMENT BY 1),
 * 	NAME CHAR(35) NOT NULL,
 * 	COUNTRY_CODE CHAR(3) NOT NULL,
 * 	DISTRICT CHAR(20) NOT NULL,
 * 	POPULATION INTEGER NOT NULL,
 * 	CONSTRAINT PK_CITY PRIMARY KEY (ID)
 * );
 * </pre>
 * 
 * @author Jonathan Earl
 * @version 1.0
 */
@Entity
@NamedQueries({
    @NamedQuery(name = "City.findAll",
                query = "SELECT c FROM City c"),
    @NamedQuery(name = "City.findByCountry",
                query = "SELECT c FROM City c WHERE c.myCountryCode = :code"),
    @NamedQuery(name = "City.findByName",
                query = "SELECT c FROM City c WHERE c.myName = :name")   
}) 
public final class City implements Serializable
{
	private static final long serialVersionUID = 1L;
	
	private static final Logger LOG = LogManager.getLogger();

	@Id
    @Column(name="ID")
    @GeneratedValue(strategy=GenerationType.IDENTITY)
	private int myId;
	@Column(name="NAME", columnDefinition="CHAR(35)")
	private String myName;
	@Column(name="COUNTRY_CODE")
	private String myCountryCode;
	@Column(name="DISTRICT")
	private String myDistrict;
	@Column(name="POPULATION")
	private int myPopulation;
	
	@Transient
	private XssSanitizerImpl sanitizer;

	/**
	 * The default constructor for the City class.
	 * <p>
	 * The initial values are:
	 * <ul>
	 *   <li>id: 1</li>
	 *   <li>name: Kabul</li>
	 *   <li>countryCode: AFG</li>
	 *   <li>population: 1780000</li>
	 * </ul>
	 */
	public City()
	{
		LOG.debug("Starting the default Constructor");
		final int initialId = 1;
		final String initalName = "Kabul";
		final String initialCountryCode = "AFG";
		final String initialDistrict = "Kabol";
		final int initialPopulation = 1780000;
		
		sanitizer = new XssSanitizerImpl();
						
		setId(initialId);
		setName(initalName);
		setCountryCode(initialCountryCode);
		setDistrict(initialDistrict);
		setPopulation(initialPopulation);
	}

	/**
	 * Returns the id value for the City.
	 *  
	 * @return the id value for the city
	 */
	public int getId()
	{
		LOG.debug("returning the Id: " + myId);
		return myId;
	}

	/**
	 * Sets the id value for the City.
	 * <p>
	 * The business rules are:
	 * <ul>
	 *   <li>the id must be 1 or greater</li>
	 * </ul>
	 * 
	 * @param id the value to set into the city id field
	 * @throws IllegalArgumentException if the id is invalid
	 */
	public void setId(final int id)
	{
		LOG.debug("setting the Id: " + id);
		final int max = 1;
		
		if (id < max)
		{
			LOG.error("Id must be greater then zero");
			throw new IllegalArgumentException("Id must be greater then zero");
		}
		this.myId = id;
	}

	/**
	 * Returns the name value for the City.
	 *  
	 * @return the name value for the city
	 */
	public String getName()
	{
		LOG.debug("returning the Name: " + myName);
		return myName;
	}

	/**
	 * Sets the name value for the City.
	 * <p>
	 * The business rules are:
	 * <ul>
	 *   <li>the name must <strong>not</strong> be null</li>
	 *   <li>the name must <strong>not</strong> be empty</li>
	 *   <li>the name must max length of 35 chars</li>
	 *   <li>XSS strings within the name will be removed</li>
	 * </ul>
	 * 
	 * @param name the value to set into the city name field
	 * @throws IllegalArgumentException if the name is invalid
	 */
	public void setName(final String name)
	{
		LOG.debug("setting the Name");
		final int max = 35;
		
		if (name == null)
		{
			LOG.error("Name must not be null");
			throw new IllegalArgumentException("Name must not be null");
		}
		
		String safeName = sanitizer.sanitizeInput(name);
		if (safeName.isEmpty())
		{
			LOG.error("Name must not be empty");
			throw new IllegalArgumentException("Name must not be empty");
		}
		if (safeName.length() > max)
		{
			LOG.error("Name must be up to 35 chars in length");
			throw new IllegalArgumentException("Name must be up to 35 chars in length");
		}
		LOG.debug("setting the Name to: " + safeName);
		this.myName = safeName;
	}

	/**
	 * Returns the countryCode value for the City.
	 *  
	 * @return the countryCode value for the city
	 */
	public String getCountryCode()
	{
		LOG.debug("returning the CountryCode: " + myCountryCode);
		return myCountryCode;
	}

	/**
	 * Sets the countryCode value for the City.
	 * <p>
	 * The business rules are:
	 * <ul>
	 *   <li>the countryCode must <strong>not</strong> be null</li>
	 *   <li>the countryCode must be 3 chars in length</li>
	 * </ul>
	 * <p>
	 * The countryCode will be converted to upper case.
	 * 
	 * @param countryCode the value to set into the city countryCode field
	 * @throws IllegalArgumentException if the countryCode is invalid
	 */
	public void setCountryCode(final String countryCode)
	{
		final int max = 3;
		LOG.debug("setting the CountryCode");
		
		if (countryCode == null)
		{
			LOG.error("CountryCode must not be null");
			throw new IllegalArgumentException("CountryCode must not be null");
		}
		String tempCountryCode = countryCode.trim().toUpperCase();
		if (tempCountryCode.length() != max)
		{
			LOG.error("CountryCode must 3 chars in length");
			throw new IllegalArgumentException("CountryCode must 3 chars in length");
		}
		LOG.debug("setting the CountryCode to: " + tempCountryCode);
		this.myCountryCode = tempCountryCode;
	}

	/**
	 * Returns the district value for the City.
	 *  
	 * @return the district value for the city
	 */
	public String getDistrict()
	{
		LOG.debug("returning the District: " + myDistrict);
		return myDistrict;
	}

	/**
	 * Sets the district value for the City.
	 * <p>
	 * The business rules are:
	 * <ul>
	 *   <li>the district must <strong>not</strong> be null</li>
	 *   <li>the district must <strong>not</strong> be empty</li>
	 *   <li>the district must max length of 20 chars</li>
	 * </ul>
	 * <p>
	 * 
	 * @param district the value to set into the city district field
	 * @throws IllegalArgumentException if the district is invalid
	 */
	public void setDistrict(final String district)
	{
		final int max = 20;
		LOG.debug("setting the District");
		
		if (district == null)
		{
			LOG.error("District must not be null");
			throw new IllegalArgumentException("District must not be null");
		}
		String tempDistrict = district.trim();
		if (tempDistrict.isEmpty())
		{
			LOG.error("District must not be empty");
			throw new IllegalArgumentException("District must not be empty");
		}
		if (tempDistrict.length() > max)
		{
			LOG.error("District must be up to 20 chars in length");
			throw new IllegalArgumentException("District must be up to 20 chars in length");
		}
		LOG.debug("setting the District to: " + tempDistrict);
		this.myDistrict = tempDistrict;
	}

	/**
	 * Returns the population value for the City.
	 *  
	 * @return the population value for the city
	 */
	public int getPopulation()
	{
		LOG.debug("returning the Population: " + myPopulation);
		return myPopulation;
	}

	/**
	 * Sets the population value for the City.
	 * <p>
	 * The business rules are:
	 * <ul>
	 *   <li>the population must be 1 or greater</li>
	 * </ul>
	 * 
	 * @param population the value to set into the city id field
	 * @throws IllegalArgumentException if the population is invalid
	 */
	public void setPopulation(final int population)
	{
		final int max = 1;
		LOG.debug("setting the Population to: " + population);
		
		if (population < max)
		{
			LOG.error("Population must be greater then zero");
			throw new IllegalArgumentException("Population must be greater then zero");
		}
		this.myPopulation = population;
	}

	/**
	 * The hashCode() method of the City class.
     * <p>
     * <strong>This method uses:</strong>
     * <ul>
     *  <li>id</li>
     *  <li>name</li>
     *  <li>countryCode</li>
     *  <li>district</li>
     *  <li>population</li>
     * </ul>
     * 
	 * @see java.lang.Object#hashCode()
	 * @return the hashCode value for this City object
	 */
	@Override
	public int hashCode()
	{
		LOG.debug("building HashCode");
		return new HashCodeBuilder()
				.append(myId)
				.append(myName)
				.append(myCountryCode)
				.append(myDistrict)
				.append(myPopulation)
				.toHashCode();
	}

	/**
	 * The equals() method of the City class.
     * <p>
     * <strong>This method uses:</strong>
     * <ul>
     *  <li>id</li>
     *  <li>name</li>
     *  <li>countryCode</li>
     *  <li>district</li>
     *  <li>population</li>
     * </ul>
     * 
	 * @see java.lang.Object#equals(Object obj)
	 * @param obj the incoming object to compare against
	 * @return true if the fields being compared are equal
	 */
	@Override
	public boolean equals(final Object obj)
	{
		LOG.debug("checking equals");
		if (obj instanceof City)
		{
			final City other = (City) obj;
			return new EqualsBuilder()
					.append(myId, other.myId)
					.append(myName, other.myName)
					.append(myCountryCode, other.myCountryCode)
					.append(myDistrict, other.myDistrict)
					.append(myPopulation, other.myPopulation)
					.isEquals();
		}
		else
		{
			return false;
		}
	}

	/**
     * The toString() method for the City class.
     * <br>
     * Outputs: City [id=XXX, name=XXX, countryCode=XXX, 
     *                district=XXX, population=XXX]
     * 
     * @see java.lang.Object#toString()
     * @return the City as a String
     */
	@Override
	public String toString()
	{
		LOG.debug("building String");
		return "City [id=" + myId + ", name=" + myName + ", countryCode=" 
	                 + myCountryCode + ", district=" + myDistrict
				     + ", population=" + myPopulation + "]";
	}
}
