package org.example.worlddao;

/*
 * This is free and unencumbered software released into the public domain.
 * Anyone is free to copy, modify, publish, use, compile, sell, or distribute this software, 
 * either in source code form or as a compiled binary, for any purpose, commercial or 
 * non-commercial, and by any means.
 * 
 * In jurisdictions that recognize copyright laws, the author or authors of this 
 * software dedicate any and all copyright interest in the software to the public domain. 
 * We make this dedication for the benefit of the public at large and to the detriment of 
 * our heirs and successors. We intend this dedication to be an overt act of relinquishment in 
 * perpetuity of all present and future rights to this software under copyright law.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * For more information, please refer to: https://unlicense.org/
*/

import java.util.List;

import javax.persistence.EntityManager;

import org.example.world.City;
import org.example.world.Country;
import org.example.worlddao.utility.NonDeleteableEntityException;
import org.example.websecurity.UserCredentials;

/**
 * This is the CityDAO Interface for the City DAO component of the World Web Application.
 * This will be the primary City database exposure for the World Web Layer.
 * 
 * @author Jonathan Earl
 * @since 1.0
 * 
 */
public interface CityDAO
{
    /**
     * Sets the JPA Entity Manager for this DAO Implication.
     * 
     * @param entityManager the Entity Manager to set
     */
    void setEntityManager(EntityManager entityManager);
        
    /**
     * Finds all Cities.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @return a List of City objects, this list maybe empty
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    List<City> findAllCities(UserCredentials credentials);    

    /**
     * Find City by Id.
     * 
     * This will return null if the City cannot be found.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param id the id for the city being requested
     * @return the City requested or null if the City cannot be found
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    City findCityById(UserCredentials credentials, int id);
    
    /**
     * Find City by Name.
     * 
     * This will return null if the City cannot be found.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param name the name for the city being requested
     * @return a List of City objects, this list maybe empty
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    List<City> findCityByName(UserCredentials credentials, String name);
    
    /**
     * Finds all Cities for a specific Country.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param countryCode the Country Code to search by
     * @return a List of City objects, this list maybe empty
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    List<City> findCitiesByCountry(UserCredentials credentials, String countryCode);    
    
    /**
     * Finds all Cities for a specific Country.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param country the Country to search by
     * @return a List of City objects, this list maybe empty
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    List<City> findCitiesByCountry(UserCredentials credentials, Country country);   
    
    /**
     * Adds a City.
     * <p>
     * Only a manager has the correct permissions for this operation.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param city the City to add
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    void addCity(UserCredentials credentials, City city);
    
    /**
     * Merge a City changed off-line back into the database.
     * <p>
     * Only a manager has the correct permissions for this operation.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param city the City to merge
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    void mergeCity(UserCredentials credentials, City city);
    
    /**
     * Check to see if a City is deleteable.
     * <p>
     * A City that is a capital of a Country is not deleteable.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param city the City to check
     * @return true if the City is deleteable
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     */
    boolean isDeleteable(UserCredentials credentials, City city);
    
    /**
     * Delete a City.
     * <p>
     * Only a manager has the correct permissions for this operation.
     * 
     * @param credentials the UserCredentials for this Database operation
     * @param city the City to merge
     * @throws IllegalArgumentException 
     *              if there is a connection or permission problem
     * @throws NonDeleteableEntityException if the City is a Capital
     *               of a Country
     */
    void deleteCity(UserCredentials credentials, City city)
    	throws NonDeleteableEntityException;
  
}
