package org.example.worlddao;

/*
 * This is free and unencumbered software released into the public domain.
 * Anyone is free to copy, modify, publish, use, compile, sell, or distribute this software, 
 * either in source code form or as a compiled binary, for any purpose, commercial or 
 * non-commercial, and by any means.
 * 
 * In jurisdictions that recognize copyright laws, the author or authors of this 
 * software dedicate any and all copyright interest in the software to the public domain. 
 * We make this dedication for the benefit of the public at large and to the detriment of 
 * our heirs and successors. We intend this dedication to be an overt act of relinquishment in 
 * perpetuity of all present and future rights to this software under copyright law.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * For more information, please refer to: https://unlicense.org/
*/

import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.Query;
import javax.persistence.TypedQuery;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.example.websecurity.UserCredentials;
import org.example.world.City;
import org.example.world.Country;
import org.example.worlddao.utility.NonDeleteableEntityException;

/**
 * This is the CityDAO Implementation for the City DAO component of the World Web Application.
 * This will be the primary City database exposure for the World Web Layer.
 * 
 * @author Jonathan Earl
 * @since 1.0
 * 
 */
public class CityDAOImpl implements CityDAO
{
    private static final Logger LOG = LogManager.getLogger();
    private EntityManager em;

    /**
     * {@inheritDoc}
     */
    @Override
    public void setEntityManager(final EntityManager entityManager)
    {
        this.em = entityManager;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<City> findAllCities(final UserCredentials credentials)
    {
        if (credentials == null || !credentials.hasRole("worker"))
        {
            LOG.error("Unauthorized access");
            throw new IllegalArgumentException("Permission refused for this operation");
        }
        TypedQuery<City> q = em.createNamedQuery("City.findAll", City.class);
        List<City> results = q.getResultList();
        return results; 
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public City findCityById(final UserCredentials credentials, final int id)
    {
    	if (credentials == null || !credentials.hasRole("worker"))
        {
            LOG.error("Unauthorized access");
            throw new IllegalArgumentException("Permission refused for this operation");
        }
        return em.find(City.class, id);
    }
    
    /**
     * {@inheritDoc}
     */
	@Override
	public List<City> findCityByName(final UserCredentials credentials, final String name) 
	{
		if (credentials == null || !credentials.hasRole("worker"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (name == null)
		{
			LOG.error("Null Name");
			throw new IllegalArgumentException("A Name must be provided");
		}
		
		TypedQuery<City> q = em.createNamedQuery("City.findByName", City.class);
		q.setParameter("name", name);
		List<City> results = q.getResultList();
		return results; 
	}


    /**
     * {@inheritDoc}
     */
	@Override
	public List<City> findCitiesByCountry(final UserCredentials credentials, final String countryCode)
	{
		if (credentials == null || !credentials.hasRole("worker"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (countryCode == null)
		{
			LOG.error("Null Country Code");
			throw new IllegalArgumentException("A Country Code must be provided");
		}
		TypedQuery<City> q = em.createNamedQuery("City.findByCountry", City.class);
		q.setParameter("code", countryCode);
		List<City> results = q.getResultList();
		return results; 
	}

	/**
     * {@inheritDoc}
     */
	@Override
	public List<City> findCitiesByCountry(final UserCredentials credentials, final Country country) 
	{
		if (credentials == null || !credentials.hasRole("worker"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (country == null)
		{
			LOG.error("Null Country");
			throw new IllegalArgumentException("A Country must be provided");
		}
		
		String code = country.getCode();
		return findCitiesByCountry(credentials, code); 
	}

	/**
     * {@inheritDoc}
     */
	@Override
	public void addCity(final UserCredentials credentials, final City city)
	{
		if (credentials == null || !credentials.hasRole("manager"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (city == null)
		{
			LOG.error("Null City");
			throw new IllegalArgumentException("A City must be provided");
		}
		em.getTransaction().begin();
		em.persist(city);
		em.getTransaction().commit();
	}

	/**
     * {@inheritDoc}
     */
	@Override
	public boolean isDeleteable(final UserCredentials credentials, final City city)
	{
		if (credentials == null || !credentials.hasRole("worker"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (city == null)
		{
			LOG.error("Null City");
			throw new IllegalArgumentException("A City must be provided");
		}
		String jpql = "Select Count(*) From Country ctr Where ctr.myCapital = :city";
		Query query = em.createQuery(jpql);
		query.setParameter("city", city);
		long count = (long) query.getSingleResult();
		return count == 0;
	}

	/**
     * {@inheritDoc}
     */
	@Override
	public void mergeCity(final UserCredentials credentials, final City city)
	{
		if (credentials == null || !credentials.hasRole("manager"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (city == null)
		{
			LOG.error("Null City");
			throw new IllegalArgumentException("A City must be provided");
		}
		em.getTransaction().begin();
		em.merge(city);
		em.getTransaction().commit();
		
	}

	/**
     * {@inheritDoc}
     */
	@Override
	public void deleteCity(final UserCredentials credentials, final City city) 
			throws NonDeleteableEntityException 
	{
		if (credentials == null || !credentials.hasRole("manager"))
		{
			LOG.error("Unauthorized access");
			throw new IllegalArgumentException("Permission refused for this operation");
		}
		if (city == null)
		{
			LOG.error("Null City");
			throw new IllegalArgumentException("A City must be provided");
		}
		if (isDeleteable(credentials, city))
		{
			em.getTransaction().begin();
			em.remove(city);
			em.getTransaction().commit();
		}
		else
		{
			throw new NonDeleteableEntityException();
		}
	}

}
