package org.example.worlddao;

import static org.junit.Assert.assertFalse;
import static org.junit.jupiter.api.Assertions.assertAll;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;
import static org.mockito.Mockito.when;

import java.io.File;
import java.util.List;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.core.config.Configurator;
import org.example.dao.testUtils.jpa.DBUnitJPAUtility;
import org.example.websecurity.UserCredentials;
import org.example.world.City;
import org.example.world.Country;
import org.example.worlddao.utility.NonDeleteableEntityException;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

public class CityDAOFastTests 
{
    private static String persistanceUnit = "World-Test";
    private static String configDir = "src" + File.separator
            + "test" + File.separator + "resources" + File.separator + "data";
    
    private static final String SCHEMA_FILE = configDir + File.separator + "World.sql";
    private static final String DATA_FILE = configDir + File.separator + "fullDB.xml";
    
    private static DBUnitJPAUtility utility = null;
    
    @BeforeAll // this runs only once before any test
    public static void setup()
    {
        Configurator.setLevel(LogManager.getLogger(City.class).getName(), Level.WARN);
        
        try
        {
            utility = new DBUnitJPAUtility(persistanceUnit, SCHEMA_FILE, DATA_FILE);
            utility.clean();
        }
        catch (Exception e)
        {
            fail(e.getLocalizedMessage()); 
        }
    }  
    
    @AfterAll
    public static void shutdown()
    {
        utility.shutdown();
    }
    
    @Test 
    public void testNullCredentialsFindAllCities() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
         
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findAllCities(null));
    }
    
    @Test
    public void testUnauthorizedFindAllCities() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(false);
    
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findAllCities(userCredentialsMock));
    }
    
    @Test
    public void testFindAllCities() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
    
        List<City> records = handler.findAllCities(userCredentialsMock);
        assertNotNull(records);
        int expected = 4079;
        int actual = records.size();
        assertEquals(expected, actual);
    }
    
    @Test 
    public void testNullCredentialsFindCityById() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        int id = 1;
    
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCityById(null, id));
    }
    
    @Test 
    public void testUnauthorizedFindCityById() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(false);
        
        int id = 1;
    
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCityById(userCredentialsMock, id));
    }
    
    @Test
    public void testFindCityByValidId() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        int id = 1;
    
        City record = handler.findCityById(userCredentialsMock, id);
        assertNotNull(record);
        String expected = "Kabul";
        String actual = record.getName().trim();
        assertEquals(expected, actual);      
    }
    
    @Test  
    public void testNullCredentialsFindCitiesByCountry() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        String countryCode = "GBR";
           
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCitiesByCountry(null, countryCode));
       
    }
    
    @Test 
    public void testUnauthorizedFindCitiesByCountry() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(false);
    
        String countryCode = "GBR";
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCitiesByCountry(userCredentialsMock, countryCode));
    }
    
    @Test 
    public void testNullCountryCodeFindCitiesByCountry() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
    
        String countryCode = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCitiesByCountry(userCredentialsMock, countryCode));
    }
    
    @Test
    public void testFindCitiesByCountry() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        String countryCode = "GBR";

        List<City> records = handler.findCitiesByCountry(userCredentialsMock, countryCode);
        
        assertAll("findCitiesByCountry", 		
        		() -> assertNotNull(records),
        		() -> assertTrue(records.size() == 81),
        		() -> assertTrue(records.stream().anyMatch(item -> "London".equals(item.getName().trim())))
        );
    }
    
    @Test
    public void testFindCitiesByInvalidCountry() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        String countryCode = "GWW";

        List<City> records = handler.findCitiesByCountry(userCredentialsMock, countryCode);
        
        assertAll("findCitiesByInvalidCountry", 
        		() -> assertNotNull(records),
        		() -> assertTrue(records.size() == 0)
        );
    }
    
    @Test  
    public void testNullCredentialsFindCitiesByCountryObject() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        Country country = new Country();
           
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCitiesByCountry(null, country));
       
    }
    
    @Test 
    public void testUnauthorizedFindCitiesByCountryObject() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(false);
    
        Country country = new Country();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCitiesByCountry(userCredentialsMock, country));
    }
    
    @Test 
    public void testNullCountryFindCitiesByCountryObject() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
    
        Country country = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCitiesByCountry(userCredentialsMock, country));
    }
    
    @Test
    public void testFindCitiesByCountryObject() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        Country country = new Country();
        country.setCode("BRA");

        List<City> records = handler.findCitiesByCountry(userCredentialsMock, country);
      
        assertAll("findCitiesByCountry", 		
        		() -> assertNotNull(records),
        		() -> assertTrue(records.size() == 250),
        		() -> assertTrue(records.stream().anyMatch(item -> "Rio de Janeiro".equals(item.getName().trim())))
        );
    }
    
    @Test
    public void testFindCitiesByInvalidCountryObject() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        Country country = new Country();
        country.setCode("BRX");

        List<City> records = handler.findCitiesByCountry(userCredentialsMock, country);
        
        assertAll("findCitiesByInvalidCountry", 
        		() -> assertNotNull(records),
        		() -> assertTrue(records.size() == 0)
        );
    }
    
    @Test  
    public void testNullCredentialsAddCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.addCity(null, city));
    }
    
    @Test 
    public void testUnauthorizedAddCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(false);
    
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.addCity(userCredentialsMock, city));
    }
    
    @Test 
    public void testNullCityAddCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
    
        City city = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.addCity(userCredentialsMock, city));
    }
    
    @Test  
    public void testNullCredentialsIsDeleteable() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.isDeleteable(null, city));
    }
    
    @Test 
    public void testUnauthorizedIsDeleteable() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(false);
    
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.isDeleteable(userCredentialsMock, city)); 
    }
    
    @Test 
    public void testNullCityIsDeleteable() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
    
        City city = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.isDeleteable(userCredentialsMock, city));
    }
    
    @Test 
    public void testsIsDeleteable() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        String name = "Miami";
    
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        City testCity = records.get(0);
             
        assertTrue(handler.isDeleteable(userCredentialsMock, testCity));
    }
    
    @Test 
    public void testsNotIsDeleteable() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        String name = "London";
    
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        City testCity = records.get(0);
             
        assertFalse(handler.isDeleteable(userCredentialsMock, testCity));
    } 
    
    @Test 
    public void testNullCredentialsFindCityByName() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        String name = "Chicago";
    
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCityByName(null, name));
    }
    
    @Test 
    public void testUnauthorizedFindCityByName() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(false);
        
        String name = "Chicago";
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCityByName(userCredentialsMock, name));
    }
    
    @Test 
    public void testNullNameFindCityByName() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        
        String name = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.findCityByName(userCredentialsMock, name));
    }
    
    @Test
    public void testFindCityByValidName() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);

        String name = "Chicago";
    
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        City target = records.get(0);
        assertNotNull(target);
        String expected = "Chicago";
        String actual = target.getName().trim();
        assertEquals(expected, actual);      
    }
    
    @Test
    public void testFindCityByInvalidName() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);

        String name = "Birnin Zana";
    
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        assertNotNull(records);
        int expected = 0;
        int actual = records.size();
        assertEquals(expected, actual);      
    }
    
    @Test  
    public void testNullCredentialsMergeCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.mergeCity(null, city));
    }
    
    @Test 
    public void testUnauthorizedMergeCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(false);
    
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.mergeCity(userCredentialsMock, city));
    }
    
    @Test 
    public void testNullCityMergeCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
    
        City city = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.mergeCity(userCredentialsMock, city));
    }
    
    @Test  
    public void testNullCredentialsDeleteCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.deleteCity(null, city));
    }
    
    @Test 
    public void testUnauthorizedDeleteCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(false);
    
        City city = new City();
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.deleteCity(userCredentialsMock, city));
    }
    
    @Test 
    public void testNullCityDeleteCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
    
        City city = null;
        
        assertThrows(IllegalArgumentException.class, () ->
        	handler.deleteCity(userCredentialsMock, city));
    }
    
    @Test 
    public void testInvalidDeleteCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
    
        String name = "Lima";
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        City city = records.get(0);
        
        assertThrows(NonDeleteableEntityException.class, () ->
        	handler.deleteCity(userCredentialsMock, city));
    }
}
