package org.example.worlddao;

import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.fail;
import static org.mockito.Mockito.when;

import java.io.File;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.core.config.Configurator;
import org.dbunit.Assertion;
import org.dbunit.dataset.ITable;
import org.example.dao.testUtils.jpa.DBUnitJPAUtility;
import org.example.websecurity.UserCredentials;
import org.example.world.City;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.mockito.Mockito;

public class CityDAOSlowTests
{
    private static String persistanceUnit = "World-Test";
    private static String configDir = "src" + File.separator
            + "test" + File.separator + "resources" + File.separator + "data";
    
    private static final String SCHEMA_FILE = configDir + File.separator + "World.sql";
    private static final String DATA_FILE = configDir + File.separator + "fullDB.xml";
    private static final String CITY_ADD_FILE = configDir + File.separator + "addCity.xml";
    private static final String CITY_MERGE_FILE = configDir + File.separator + "mergeCity.xml";
    private static final String CITY_DELETE_FILE = configDir + File.separator + "deleteCity.xml";
       
    private static DBUnitJPAUtility utility = null;
    
    @BeforeEach // rebuild the test database before each test
    public void setup()
    {
        Configurator.setLevel(LogManager.getLogger(City.class).getName(), Level.WARN);
        
        try
        {
            utility = new DBUnitJPAUtility(persistanceUnit, SCHEMA_FILE, DATA_FILE);
            utility.clean();
        }
        catch (Exception e)
        { 
            fail(e.getLocalizedMessage());
        }
    }  
    
    @AfterEach
    public void shutdown()
    {
        utility.shutdown();
    }
    
    @Timeout(3) // 3 seconds timeout
    @Test
    public void testAddCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
        
        City country = buildTestCity();
     
        handler.addCity(userCredentialsMock, country);
        
        try
        {
            // Fetch database data after executing your code
            ITable actualTable = utility.getTableFromDatabase("CITY");
            ITable expectedTable = utility.getTableFromFile(CITY_ADD_FILE, "CITY");

            // Assert actual database table match expected table
            // This will check every row, and every column of the table
            Assertion.assertEquals(expectedTable, actualTable);
        }
        catch (Exception e1)
        {
            fail(e1.getLocalizedMessage());
        }
    }
    
    @Timeout(value = 3, unit = TimeUnit.SECONDS)  
    @Test
    public void testMergeCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
        
        String name = "Detroit";
        
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        City city = records.get(0);
        utility.getEntityManager().detach(city);  // release it from the EntityManger
        
        city.setPopulation(1000000);  // at this point the database record does not match the record in memory
        
        List<City> records2 = handler.findCityByName(userCredentialsMock, name);
        City city2 = records2.get(0);
        
        assertFalse(city.equals(city2));
     
        handler.mergeCity(userCredentialsMock, city);  // update the database w/ the new values
        
        try
        {
            // Fetch database data after executing your code
            ITable actualTable = utility.getTableFromDatabase("CITY");
            ITable expectedTable = utility.getTableFromFile(CITY_MERGE_FILE, "CITY");

            // Assert actual database table match expected table
            // This will check every row, and every column of the table
            Assertion.assertEquals(expectedTable, actualTable);
        }
        catch (Exception e1)
        {
            fail(e1.getLocalizedMessage());
        }
    }
    
    @Timeout(value = 3, unit = TimeUnit.SECONDS)  
    @Test
    public void testDeleteCity() 
            throws Exception
    {
        CityDAO handler = new CityDAOImpl();
        handler.setEntityManager(utility.getEntityManager());
        
        UserCredentials userCredentialsMock = Mockito.mock(UserCredentials.class);
        when(userCredentialsMock.hasRole("worker")).thenReturn(true);
        when(userCredentialsMock.hasRole("manager")).thenReturn(true);
        
        String name = "Boulder";
        
        List<City> records = handler.findCityByName(userCredentialsMock, name);
        City city = records.get(0);
        handler.deleteCity(userCredentialsMock, city);
           
        try
        {
            // Fetch database data after executing your code
            ITable actualTable = utility.getTableFromDatabase("CITY");
            ITable expectedTable = utility.getTableFromFile(CITY_DELETE_FILE, "CITY");

            // Assert actual database table match expected table
            // This will check every row, and every column of the table
            Assertion.assertEquals(expectedTable, actualTable);
        }
        catch (Exception e1)
        {
            fail(e1.getLocalizedMessage());
        }
    }
    
    private City buildTestCity()
    {
    	City current = new City();
   // 	current.setId(4080);   // I cannot set this or JPA thinks the record already exists
    	current.setName("Shinjuku");
    	current.setCountryCode("JPN");
    	current.setDistrict("Kantō");
    	current.setPopulation(34235);
    	return current;
    }
}
